@@ -1,16 +1,19 @@
 # Copyright (C) 2015  JWCrypto Project Contributors - see LICENSE file
 
 import os
-
 from binascii import hexlify, unhexlify
+from collections import namedtuple
+from enum import Enum
 
+from cryptography import x509
 from cryptography.hazmat.backends import default_backend
-from cryptography.hazmat.primitives import hashes
+from cryptography.hazmat.primitives import hashes, serialization
 from cryptography.hazmat.primitives.asymmetric import ec
 from cryptography.hazmat.primitives.asymmetric import rsa
 
 from six import iteritems
 
+from jwcrypto.common import JWException
 from jwcrypto.common import base64url_decode, base64url_encode
 from jwcrypto.common import json_decode, json_encode
 
@@ -21,36 +24,59 @@ JWKTypesRegistry = {'EC': 'Elliptic Curve',
                     'oct': 'Octet sequence'}
 """Registry of valid Key Types"""
 
+
 # RFC 7518 - 7.5
 # It is part of the JWK Parameters Registry, but we want a more
 # specific map for internal usage
-JWKValuesRegistry = {'EC': {'crv': ('Curve', 'Public', 'Required'),
-                            'x': ('X Coordinate', 'Public', 'Required'),
-                            'y': ('Y Coordinate', 'Public', 'Required'),
-                            'd': ('ECC Private Key', 'Private', None)},
-                     'RSA': {'n': ('Modulus', 'Public', 'Required'),
-                             'e': ('Exponent', 'Public', 'Required'),
-                             'd': ('Private Exponent', 'Private', None),
-                             'p': ('First Prime Factor', 'Private', None),
-                             'q': ('Second Prime Factor', 'Private', None),
-                             'dp': ('First Factor CRT Exponent', 'Private',
-                                    None),
-                             'dq': ('Second Factor CRT Exponent', 'Private',
-                                    None),
-                             'qi': ('First CRT Coefficient', 'Private', None)},
-                     'oct': {'k': ('Key Value', 'Private', 'Required')}}
+class ParmType(Enum):
+    name = 'A string with a name'
+    b64 = 'Base64url Encoded'
+    b64U = 'Base64urlUint Encoded'
+    unsupported = 'Unsupported Parameter'
+
+
+JWKParameter = namedtuple('Parameter', 'description public required type')
+JWKValuesRegistry = {
+    'EC': {
+        'crv': JWKParameter('Curve', True, True, ParmType.name),
+        'x': JWKParameter('X Coordinate', True, True, ParmType.b64),
+        'y': JWKParameter('Y Coordinate', True, True, ParmType.b64),
+        'd': JWKParameter('ECC Private Key', False, False, ParmType.b64),
+    },
+    'RSA': {
+        'n': JWKParameter('Modulus', True, True, ParmType.b64),
+        'e': JWKParameter('Exponent', True, True, ParmType.b64U),
+        'd': JWKParameter('Private Exponent', False, False, ParmType.b64U),
+        'p': JWKParameter('First Prime Factor', False, False, ParmType.b64U),
+        'q': JWKParameter('Second Prime Factor', False, False, ParmType.b64U),
+        'dp': JWKParameter('First Factor CRT Exponent',
+                           False, False, ParmType.b64U),
+        'dq': JWKParameter('Second Factor CRT Exponent',
+                           False, False, ParmType.b64U),
+        'qi': JWKParameter('First CRT Coefficient',
+                           False, False, ParmType.b64U),
+        'oth': JWKParameter('Other Primes Info',
+                            False, False, ParmType.unsupported),
+    },
+    'oct': {
+        'k': JWKParameter('Key Value', False, True, ParmType.b64),
+    }
+}
 """Registry of valid key values"""
 
-JWKParamsRegistry = {'kty': ('Key Type', 'Public', ),
-                     'use': ('Public Key Use', 'Public'),
-                     'key_ops': ('Key Operations', 'Public'),
-                     'alg': ('Algorithm', 'Public'),
-                     'kid': ('Key ID', 'Public'),
-                     'x5u': ('X.509 URL', 'Public'),
-                     'x5c': ('X.509 Certificate Chain', 'Public'),
-                     'x5t': ('X.509 Certificate SHA-1 Thumbprint', 'Public'),
-                     'x5t#S256': ('X.509 Certificate SHA-256 Thumbprint',
-                                  'Public')}
+JWKParamsRegistry = {
+    'kty': JWKParameter('Key Type', True, None, None),
+    'use': JWKParameter('Public Key Use', True, None, None),
+    'key_ops': JWKParameter('Key Operations', True, None, None),
+    'alg': JWKParameter('Algorithm', True, None, None),
+    'kid': JWKParameter('Key ID', True, None, None),
+    'x5u': JWKParameter('X.509 URL', True, None, None),
+    'x5c': JWKParameter('X.509 Certificate Chain', True, None, None),
+    'x5t': JWKParameter('X.509 Certificate SHA-1 Thumbprint',
+                        True, None, None),
+    'x5t#S256': JWKParameter('X.509 Certificate SHA-256 Thumbprint',
+                             True, None, None)
+}
 """Regstry of valid key parameters"""
 
 # RFC 7518 - 7.6
@@ -82,7 +108,7 @@ JWKpycaCurveMap = {'secp256r1': 'P-256',
                    'secp521r1': 'P-521'}
 
 
-class InvalidJWKType(Exception):
+class InvalidJWKType(JWException):
     """Invalid JWK Type Exception.
 
     This exception is raised when an invalid parameter type is used.
@@ -97,7 +123,7 @@ class InvalidJWKType(Exception):
             self.value, list(JWKTypesRegistry.keys()))
 
 
-class InvalidJWKUsage(Exception):
+class InvalidJWKUsage(JWException):
     """Invalid JWK usage Exception.
 
     This exception is raised when an invalid key usage is requested,
@@ -122,7 +148,7 @@ class InvalidJWKUsage(Exception):
                                                                    valid)
 
 
-class InvalidJWKOperation(Exception):
+class InvalidJWKOperation(JWException):
     """Invalid JWK Operation Exception.
 
     This exception is raised when an invalid key operation is requested,
@@ -149,7 +175,7 @@ class InvalidJWKOperation(Exception):
                                                                        valid)
 
 
-class InvalidJWKValue(Exception):
+class InvalidJWKValue(JWException):
     """Invalid JWK Value Exception.
 
     This exception is raised when an invalid/unknown value is used in the
@@ -165,7 +191,7 @@ class JWK(object):
 
     This object represent a Key.
     It must be instantiated by using the standard defined key/value pairs
-    as arguents of the initialization function.
+    as arguments of the initialization function.
     """
 
     def __init__(self, **kwargs):
@@ -209,6 +235,7 @@ class JWK(object):
     @classmethod
     def generate(cls, **kwargs):
         obj = cls()
+        kty = None
         try:
             kty = kwargs['kty']
             gen = getattr(obj, '_generate_%s' % kty)
@@ -218,20 +245,30 @@ class JWK(object):
         return obj
 
     def generate_key(self, **params):
+        kty = None
         try:
-            kty = params['generate']
-            del params['generate']
+            kty = params.pop('generate')
             gen = getattr(self, '_generate_%s' % kty)
         except (KeyError, AttributeError):
             raise InvalidJWKType(kty)
 
         gen(params)
 
-    def _generate_oct(self, params):
-        size = 128
+    def _get_gen_size(self, params, default_size=None):
+        size = default_size
         if 'size' in params:
-            size = params['size']
-            del params['size']
+            size = params.pop('size')
+        elif 'alg' in params:
+            try:
+                from jwcrypto.jwa import JWA
+                alg = JWA.instantiate_alg(params['alg'])
+            except KeyError:
+                raise ValueError("Invalid 'alg' parameter")
+            size = alg.keysize
+        return size
+
+    def _generate_oct(self, params):
+        size = self._get_gen_size(params, 128)
         key = os.urandom(size // 8)
         params['kty'] = 'oct'
         params['k'] = base64url_encode(key)
@@ -243,13 +280,9 @@ class JWK(object):
 
     def _generate_RSA(self, params):
         pubexp = 65537
-        size = 2048
+        size = self._get_gen_size(params, 2048)
         if 'public_exponent' in params:
-            pubexp = params['public_exponent']
-            del params['public_exponent']
-        if 'size' in params:
-            size = params['size']
-            del params['size']
+            pubexp = params.pop('public_exponent')
         key = rsa.generate_private_key(pubexp, size, default_backend())
         self._import_pyca_pri_rsa(key, **params)
 
@@ -290,13 +323,11 @@ class JWK(object):
     def _generate_EC(self, params):
         curve = 'P-256'
         if 'curve' in params:
-            curve = params['curve']
-            del params['curve']
+            curve = params.pop('curve')
         # 'curve' is for backwards compat, if 'crv' is defined it takes
         # precedence
         if 'crv' in params:
-            curve = params['crv']
-            del params['crv']
+            curve = params.pop('crv')
         curve_name = self._get_curve_by_name(curve)
         key = ec.generate_private_key(curve_name, default_backend())
         self._import_pyca_pri_ec(key, **params)
@@ -342,8 +373,26 @@ class JWK(object):
                     names.remove(name)
 
         for name, val in iteritems(JWKValuesRegistry[kty]):
-            if val[2] == 'Required' and name not in self._key:
+            if val.required and name not in self._key:
                 raise InvalidJWKValue('Missing required value %s' % name)
+            if val.type == ParmType.unsupported and name in self._key:
+                raise InvalidJWKValue('Unsupported parameter %s' % name)
+            if val.type == ParmType.b64 and name in self._key:
+                # Check that the value is base64url encoded
+                try:
+                    base64url_decode(self._key[name])
+                except Exception:  # pylint: disable=broad-except
+                    raise InvalidJWKValue(
+                        '"%s" is not base64url encoded' % name
+                    )
+            if val[3] == ParmType.b64U and name in self._key:
+                # Check that the value is Base64urlUInt encoded
+                try:
+                    self._decode_int(self._key[name])
+                except Exception:  # pylint: disable=broad-except
+                    raise InvalidJWKValue(
+                        '"%s" is not Base64urlUInt encoded' % name
+                    )
 
         # Unknown key parameters are allowed
         # Let's just store them out of the way
@@ -381,36 +430,107 @@ class JWK(object):
                                               ' "key_ops" values specified at'
                                               ' the same time')
 
+    @classmethod
+    def from_json(cls, key):
+        """Creates a RFC 7517 JWK from the standard JSON format.
+
+        :param key: The RFC 7517 representation of a JWK.
+        """
+        obj = cls()
+        try:
+            jkey = json_decode(key)
+        except Exception as e:  # pylint: disable=broad-except
+            raise InvalidJWKValue(e)
+        obj.import_key(**jkey)
+        return obj
+
     def export(self, private_key=True):
         """Exports the key in the standard JSON format.
+        Exports the key regardless of type, if private_key is False
+        and the key is_symmetric an exceptionis raised.
 
         :param private_key(bool): Whether to export the private key.
                                   Defaults to True.
         """
-        if private_key is not True:
+        if private_key is True:
+            # Use _export_all for backwards compatibility, as this
+            # function allows to export symmetrict keys too
+            return self._export_all()
+        else:
             return self.export_public()
+
+    def export_public(self):
+        """Exports the public key in the standard JSON format.
+        It fails if one is not available like when this function
+        is called on a symmetric key.
+        """
+        pub = self._public_params()
+        return json_encode(pub)
+
+    def _public_params(self):
+        if not self.has_public:
+            raise InvalidJWKType("No public key available")
+        pub = {}
+        preg = JWKParamsRegistry
+        for name in preg:
+            if preg[name].public:
+                if name in self._params:
+                    pub[name] = self._params[name]
+        reg = JWKValuesRegistry[self._params['kty']]
+        for param in reg:
+            if reg[param].public:
+                pub[param] = self._key[param]
+        return pub
+
+    def _export_all(self):
         d = dict()
         d.update(self._params)
         d.update(self._key)
         d.update(self._unknown)
         return json_encode(d)
 
-    def export_public(self):
-        """Exports the public key in the standard JSON format.
-           This function is deprecated and maintained only for
-           backwards compatibility, use export(private_key=False)
-           instead."""
-        pub = {}
-        preg = JWKParamsRegistry
-        for name in preg:
-            if preg[name][1] == 'Public':
-                if name in self._params:
-                    pub[name] = self._params[name]
+    def export_private(self):
+        """Export the private key in the standard JSON format.
+        It fails for a JWK that has only a public key or is symmetric.
+        """
+        if self.has_private:
+            return self._export_all()
+        raise InvalidJWKType("No private key available")
+
+    def export_symmetric(self):
+        if self.is_symmetric:
+            return self._export_all()
+        raise InvalidJWKType("Not a symmetric key")
+
+    def public(self):
+        pub = self._public_params()
+        return JWK(**pub)
+
+    @property
+    def has_public(self):
+        """Whether this JWK has an asymmetric Public key."""
+        if self.is_symmetric:
+            return False
         reg = JWKValuesRegistry[self._params['kty']]
-        for param in reg:
-            if reg[param][1] == 'Public':
-                pub[param] = self._key[param]
-        return json_encode(pub)
+        for value in reg:
+            if reg[value].public and value in self._key:
+                return True
+
+    @property
+    def has_private(self):
+        """Whether this JWK has an asymmetric key Private key."""
+        if self.is_symmetric:
+            return False
+        reg = JWKValuesRegistry[self._params['kty']]
+        for value in reg:
+            if not reg[value].public and value in self._key:
+                return True
+        return False
+
+    @property
+    def is_symmetric(self):
+        """Whether this JWK is a symmetric key."""
+        return self.key_type == 'oct'
 
     @property
     def key_type(self):
@@ -556,12 +676,89 @@ class JWK(object):
         else:
             raise InvalidJWKValue('Unknown key object %r' % key)
 
+    def import_from_pem(self, data, password=None):
+        """Imports a key from data loaded from a PEM file.
+        The key may be encrypted with a password.
+        Private keys (PKCS#8 format), public keys, and X509 certificate's
+        public keys can be imported with this interface.
+
+        :param data(bytes): The data contained in a PEM file.
+        :param password(bytes): An optional password to unwrap the key.
+        """
+
+        try:
+            key = serialization.load_pem_private_key(
+                data, password=password, backend=default_backend())
+        except ValueError as e:
+            if password is not None:
+                raise e
+            try:
+                key = serialization.load_pem_public_key(
+                    data, backend=default_backend())
+            except ValueError:
+                try:
+                    cert = x509.load_pem_x509_certificate(
+                        data, backend=default_backend())
+                    key = cert.public_key()
+                except ValueError:
+                    raise e
+
+        self.import_from_pyca(key)
+        self._params['kid'] = self.thumbprint()
+
+    def export_to_pem(self, private_key=False, password=False):
+        """Exports keys to a data buffer suitable to be stored as a PEM file.
+        Either the public or the private key can be exported to a PEM file.
+        For private keys the PKCS#8 format is used. If a password is provided
+        the best encryption method available as determined by the cryptography
+        module is used to wrap the key.
+
+        :param private_key: Whether the private key should be exported.
+         Defaults to `False` which means the public key is exported by default.
+        :param password(bytes): A password for wrapping the private key.
+         Defaults to False which will cause the operation to fail. To avoid
+         encryption the user must explicitly pass None, otherwise the user
+         needs to provide a password in a bytes buffer.
+        """
+        e = serialization.Encoding.PEM
+        if private_key:
+            if not self.has_private:
+                raise InvalidJWKType("No private key available")
+            f = serialization.PrivateFormat.PKCS8
+            if password is None:
+                a = serialization.NoEncryption()
+            elif isinstance(password, bytes):
+                a = serialization.BestAvailableEncryption(password)
+            elif password is False:
+                raise ValueError("The password must be None or a bytes string")
+            else:
+                raise TypeError("The password string must be bytes")
+            return self._get_private_key().private_bytes(
+                encoding=e, format=f, encryption_algorithm=a)
+        else:
+            if not self.has_public:
+                raise InvalidJWKType("No public key available")
+            f = serialization.PublicFormat.SubjectPublicKeyInfo
+            return self._get_public_key().public_bytes(encoding=e, format=f)
+
     @classmethod
     def from_pyca(cls, key):
         obj = cls()
         obj.import_from_pyca(key)
         return obj
 
+    @classmethod
+    def from_pem(cls, data, password=None):
+        """Creates a key from PKCS#8 formatted data loaded from a PEM file.
+           See the function `import_from_pem` for details.
+
+        :param data(bytes): The data contained in a PEM file.
+        :param password(bytes): An optional password to unwrap the key.
+        """
+        obj = cls()
+        obj.import_from_pem(data, password)
+        return obj
+
     def thumbprint(self, hashalg=hashes.SHA256()):
         """Returns the key thumbprint as specified by RFC 7638.
 
@@ -570,7 +767,7 @@ class JWK(object):
 
         t = {'kty': self._params['kty']}
         for name, val in iteritems(JWKValuesRegistry[t['kty']]):
-            if val[2] == 'Required':
+            if val.required:
                 t[name] = self._key[name]
         digest = hashes.Hash(hashalg, backend=default_backend())
         digest.update(bytes(json_encode(t).encode('utf8')))
@@ -603,6 +800,12 @@ class JWKSet(dict):
         super(JWKSet, self).__setitem__('keys', _JWKkeys())
         self.update(*args, **kwargs)
 
+    def __iter__(self):
+        return self['keys'].__iter__()
+
+    def __contains__(self, key):
+        return self['keys'].__contains__(key)
+
     def __setitem__(self, key, val):
         if key == 'keys':
             self['keys'].add(val)
@@ -639,7 +842,7 @@ class JWKSet(dict):
         """
         try:
             jwkset = json_decode(keyset)
-        except:
+        except Exception:  # pylint: disable=broad-except
             raise InvalidJWKValue()
 
         if 'keys' not in jwkset:
@@ -652,8 +855,6 @@ class JWKSet(dict):
             else:
                 self[k] = v
 
-        return self
-
     @classmethod
     def from_json(cls, keyset):
         """Creates a RFC 7517 keyset from the standard JSON format.
@@ -661,7 +862,8 @@ class JWKSet(dict):
         :param keyset: The RFC 7517 representation of a JOSE Keyset.
         """
         obj = cls()
-        return obj.import_keyset(keyset)
+        obj.import_keyset(keyset)
+        return obj
 
     def get_key(self, kid):
         """Gets a key from the set.
