@@ -1,34 +1,34 @@
 # Copyright (C) 2015 JWCrypto Project Contributors - see LICENSE file
 
-from binascii import hexlify, unhexlify
+from collections import namedtuple
 
-from cryptography.exceptions import InvalidSignature
-from cryptography.hazmat.backends import default_backend
-from cryptography.hazmat.primitives import hashes, hmac
-from cryptography.hazmat.primitives.asymmetric import ec
-from cryptography.hazmat.primitives.asymmetric import padding
-from cryptography.hazmat.primitives.asymmetric import utils as ec_utils
-
-from jwcrypto.common import InvalidJWAAlgorithm
+from jwcrypto.common import JWException
 from jwcrypto.common import base64url_decode, base64url_encode
 from jwcrypto.common import json_decode, json_encode
+from jwcrypto.jwa import JWA
 from jwcrypto.jwk import JWK
 
 
 # RFC 7515 - 9.1
 # name: (description, supported?)
-JWSHeaderRegistry = {'alg': ('Algorithm', True),
-                     'jku': ('JWK Set URL', False),
-                     'jwk': ('JSON Web Key', False),
-                     'kid': ('Key ID', True),
-                     'x5u': ('X.509 URL', False),
-                     'x5c': ('X.509 Certificate Chain', False),
-                     'x5t': ('X.509 Certificate SHA-1 Thumbprint', False),
-                     'x5t#S256': ('X.509 Certificate SHA-256 Thumbprint',
-                                  False),
-                     'typ': ('Type', True),
-                     'cty': ('Content Type', True),
-                     'crit': ('Critical', True)}
+JWSHeaderParameter = namedtuple('Parameter',
+                                'description mustprotect supported')
+JWSHeaderRegistry = {
+    'alg': JWSHeaderParameter('Algorithm', False, True),
+    'jku': JWSHeaderParameter('JWK Set URL', False, False),
+    'jwk': JWSHeaderParameter('JSON Web Key', False, False),
+    'kid': JWSHeaderParameter('Key ID', False, True),
+    'x5u': JWSHeaderParameter('X.509 URL', False, False),
+    'x5c': JWSHeaderParameter('X.509 Certificate Chain', False, False),
+    'x5t': JWSHeaderParameter(
+        'X.509 Certificate SHA-1 Thumbprint', False, False),
+    'x5t#S256': JWSHeaderParameter(
+        'X.509 Certificate SHA-256 Thumbprint', False, False),
+    'typ': JWSHeaderParameter('Type', False, True),
+    'cty': JWSHeaderParameter('Content Type', False, True),
+    'crit': JWSHeaderParameter('Critical', True, True),
+    'b64': JWSHeaderParameter('Base64url-Encode Payload', True, True)
+}
 """Registry of valid header parameters"""
 
 default_allowed_algs = [
@@ -39,7 +39,7 @@ default_allowed_algs = [
 """Default allowed algorithms"""
 
 
-class InvalidJWSSignature(Exception):
+class InvalidJWSSignature(JWException):
     """Invalid JWS Signature.
 
     This exception is raised when a signature cannot be validated.
@@ -56,7 +56,7 @@ class InvalidJWSSignature(Exception):
         super(InvalidJWSSignature, self).__init__(msg)
 
 
-class InvalidJWSObject(Exception):
+class InvalidJWSObject(JWException):
     """Invalid JWS Object.
 
     This exception is raised when the JWS Object is invalid and/or
@@ -72,7 +72,7 @@ class InvalidJWSObject(Exception):
         super(InvalidJWSObject, self).__init__(msg)
 
 
-class InvalidJWSOperation(Exception):
+class InvalidJWSOperation(JWException):
     """Invalid JWS Object.
 
     This exception is raised when a requested operation cannot
@@ -90,100 +90,6 @@ class InvalidJWSOperation(Exception):
         super(InvalidJWSOperation, self).__init__(msg)
 
 
-class _RawJWS(object):
-
-    def sign(self, key, payload):
-        raise NotImplementedError
-
-    def verify(self, key, payload, signature):
-        raise NotImplementedError
-
-
-class _RawHMAC(_RawJWS):
-
-    def __init__(self, hashfn):
-        self.backend = default_backend()
-        self.hashfn = hashfn
-
-    def _hmac_setup(self, key, payload):
-        h = hmac.HMAC(key, self.hashfn, backend=self.backend)
-        h.update(payload)
-        return h
-
-    def sign(self, key, payload):
-        skey = base64url_decode(key.get_op_key('sign'))
-        h = self._hmac_setup(skey, payload)
-        return h.finalize()
-
-    def verify(self, key, payload, signature):
-        vkey = base64url_decode(key.get_op_key('verify'))
-        h = self._hmac_setup(vkey, payload)
-        try:
-            h.verify(signature)
-        except InvalidSignature as e:
-            raise InvalidJWSSignature(exception=e)
-
-
-class _RawRSA(_RawJWS):
-    def __init__(self, padfn, hashfn):
-        self.padfn = padfn
-        self.hashfn = hashfn
-
-    def sign(self, key, payload):
-        skey = key.get_op_key('sign')
-        signer = skey.signer(self.padfn, self.hashfn)
-        signer.update(payload)
-        return signer.finalize()
-
-    def verify(self, key, payload, signature):
-        pkey = key.get_op_key('verify')
-        verifier = pkey.verifier(signature, self.padfn, self.hashfn)
-        verifier.update(payload)
-        verifier.verify()
-
-
-class _RawEC(_RawJWS):
-    def __init__(self, curve, hashfn):
-        self.curve = curve
-        self.hashfn = hashfn
-
-    def encode_int(self, n, l):
-        e = hex(n).rstrip("L").lstrip("0x")
-        ilen = (l + 7) // 8  # number of bytes rounded up
-        e = '0' * (ilen * 2 - len(e)) + e  # pad as necessary
-        return unhexlify(e)
-
-    def sign(self, key, payload):
-        skey = key.get_op_key('sign', self.curve)
-        signer = skey.signer(ec.ECDSA(self.hashfn))
-        signer.update(payload)
-        signature = signer.finalize()
-        r, s = ec_utils.decode_rfc6979_signature(signature)
-        l = key.get_curve(self.curve).key_size
-        return self.encode_int(r, l) + self.encode_int(s, l)
-
-    def verify(self, key, payload, signature):
-        pkey = key.get_op_key('verify', self.curve)
-        r = signature[:len(signature) // 2]
-        s = signature[len(signature) // 2:]
-        enc_signature = ec_utils.encode_rfc6979_signature(
-            int(hexlify(r), 16), int(hexlify(s), 16))
-        verifier = pkey.verifier(enc_signature, ec.ECDSA(self.hashfn))
-        verifier.update(payload)
-        verifier.verify()
-
-
-class _RawNone(_RawJWS):
-
-    def sign(self, key, payload):
-        return ''
-
-    def verify(self, key, payload, signature):
-        if signature != b'':
-            raise InvalidJWSSignature('The "none" signature must be the '
-                                      'empty string')
-
-
 class JWSCore(object):
     """The inner JWS Core object.
 
@@ -215,74 +121,41 @@ class JWSCore(object):
         self.key = key
 
         if header is not None:
+            if isinstance(header, dict):
+                self.header = header
+                header = json_encode(header)
+            else:
+                self.header = json_decode(header)
+
             self.protected = base64url_encode(header.encode('utf-8'))
         else:
+            self.header = dict()
             self.protected = ''
-        self.payload = base64url_encode(payload)
-
-    def _jwa_HS256(self):
-        return _RawHMAC(hashes.SHA256())
-
-    def _jwa_HS384(self):
-        return _RawHMAC(hashes.SHA384())
-
-    def _jwa_HS512(self):
-        return _RawHMAC(hashes.SHA512())
-
-    def _jwa_RS256(self):
-        return _RawRSA(padding.PKCS1v15(), hashes.SHA256())
-
-    def _jwa_RS384(self):
-        return _RawRSA(padding.PKCS1v15(), hashes.SHA384())
-
-    def _jwa_RS512(self):
-        return _RawRSA(padding.PKCS1v15(), hashes.SHA512())
-
-    def _jwa_ES256(self):
-        return _RawEC('P-256', hashes.SHA256())
-
-    def _jwa_ES384(self):
-        return _RawEC('P-384', hashes.SHA384())
-
-    def _jwa_ES512(self):
-        return _RawEC('P-521', hashes.SHA512())
-
-    def _jwa_PS256(self):
-        return _RawRSA(padding.PSS(padding.MGF1(hashes.SHA256()),
-                                   hashes.SHA256.digest_size),
-                       hashes.SHA256())
-
-    def _jwa_PS384(self):
-        return _RawRSA(padding.PSS(padding.MGF1(hashes.SHA384()),
-                                   hashes.SHA384.digest_size),
-                       hashes.SHA384())
-
-    def _jwa_PS512(self):
-        return _RawRSA(padding.PSS(padding.MGF1(hashes.SHA512()),
-                                   hashes.SHA512.digest_size),
-                       hashes.SHA512())
-
-    def _jwa_none(self):
-        return _RawNone()
+        self.payload = payload
 
     def _jwa(self, name, allowed):
         if allowed is None:
             allowed = default_allowed_algs
-        attr = '_jwa_%s' % name
-        try:
-            fn = getattr(self, attr)
-        except (KeyError, AttributeError):
-            raise InvalidJWAAlgorithm()
         if name not in allowed:
             raise InvalidJWSOperation('Algorithm not allowed')
-        return fn()
+        return JWA.signing_alg(name)
+
+    def _payload(self):
+        if self.header.get('b64', True):
+            return base64url_encode(self.payload).encode('utf-8')
+        else:
+            if isinstance(self.payload, bytes):
+                return self.payload
+            else:
+                return self.payload.encode('utf-8')
 
     def sign(self):
         """Generates a signature"""
-        sigin = ('.'.join([self.protected, self.payload])).encode('utf-8')
+        payload = self._payload()
+        sigin = b'.'.join([self.protected.encode('utf-8'), payload])
         signature = self.engine.sign(self.key, sigin)
         return {'protected': self.protected,
-                'payload': self.payload,
+                'payload': payload,
                 'signature': base64url_encode(signature)}
 
     def verify(self, signature):
@@ -291,7 +164,8 @@ class JWSCore(object):
         :raises InvalidJWSSignature: if the verification fails.
         """
         try:
-            sigin = ('.'.join([self.protected, self.payload])).encode('utf-8')
+            payload = self._payload()
+            sigin = b'.'.join([self.protected.encode('utf-8'), payload])
             self.engine.verify(self.key, sigin, signature)
         except Exception as e:  # pylint: disable=broad-except
             raise InvalidJWSSignature('Verification failed', repr(e))
@@ -315,16 +189,6 @@ class JWS(object):
         self.verifylog = None
         self._allowed_algs = None
 
-    def _check_crit(self, crit):
-        for k in crit:
-            if k not in JWSHeaderRegistry:
-                raise InvalidJWSSignature('Unknown critical header: '
-                                          '"%s"' % k)
-            else:
-                if not JWSHeaderRegistry[k][1]:
-                    raise InvalidJWSSignature('Unsupported critical '
-                                              'header: "%s"' % k)
-
     @property
     def allowed_algs(self):
         """Allowed algorithms.
@@ -348,31 +212,61 @@ class JWS(object):
     def is_valid(self):
         return self.objects.get('valid', False)
 
-    def _merge_headers(self, h1, h2):
-        for k in list(h1.keys()):
-            if k in h2:
-                raise InvalidJWSObject('Duplicate header: "%s"' % k)
-        h1.update(h2)
-        return h1
+    # TODO: allow caller to specify list of headers it understands
+    def _merge_check_headers(self, protected, *headers):
+        header = None
+        crit = []
+        if protected is not None:
+            if 'crit' in protected:
+                crit = protected['crit']
+                # Check immediately if we support these critical headers
+                for k in crit:
+                    if k not in JWSHeaderRegistry:
+                        raise InvalidJWSObject(
+                            'Unknown critical header: "%s"' % k)
+                    else:
+                        if not JWSHeaderRegistry[k][1]:
+                            raise InvalidJWSObject(
+                                'Unsupported critical header: "%s"' % k)
+            header = protected
+            if 'b64' in header:
+                if not isinstance(header['b64'], bool):
+                    raise InvalidJWSObject('b64 header must be a boolean')
+
+        for hn in headers:
+            if hn is None:
+                continue
+            if header is None:
+                header = dict()
+            for h in list(hn.keys()):
+                if h in JWSHeaderRegistry:
+                    if JWSHeaderRegistry[h].mustprotect:
+                        raise InvalidJWSObject('"%s" must be protected' % h)
+                if h in header:
+                    raise InvalidJWSObject('Duplicate header: "%s"' % h)
+            header.update(hn)
+
+        for k in crit:
+            if k not in header:
+                raise InvalidJWSObject('Missing critical header "%s"' % k)
+
+        return header
 
     # TODO: support selecting key with 'kid' and passing in multiple keys
     def _verify(self, alg, key, payload, signature, protected, header=None):
-        # verify it is a valid JSON object and keep a decode copy
+        p = dict()
+        # verify it is a valid JSON object and decode
         if protected is not None:
             p = json_decode(protected)
-        else:
-            p = dict()
-        if not isinstance(p, dict):
-            raise InvalidJWSSignature('Invalid Protected header')
+            if not isinstance(p, dict):
+                raise InvalidJWSSignature('Invalid Protected header')
         # merge heders, and verify there are no duplicates
         if header:
             if not isinstance(header, dict):
                 raise InvalidJWSSignature('Invalid Unprotected header')
-            p = self._merge_headers(p, header)
-        # verify critical headers
-        # TODO: allow caller to specify list of headers it understands
-        if 'crit' in p:
-            self._check_crit(p['crit'])
+
+        # Merge and check (critical) headers
+        self._merge_check_headers(p, header)
         # check 'alg' is present
         if alg is None and 'alg' not in p:
             raise InvalidJWSSignature('No "alg" in headers')
@@ -433,6 +327,33 @@ class JWS(object):
             raise InvalidJWSSignature('Verification failed for all '
                                       'signatures' + repr(self.verifylog))
 
+    def _deserialize_signature(self, s):
+        o = dict()
+        o['signature'] = base64url_decode(str(s['signature']))
+        if 'protected' in s:
+            p = base64url_decode(str(s['protected']))
+            o['protected'] = p.decode('utf-8')
+        if 'header' in s:
+            o['header'] = s['header']
+        return o
+
+    def _deserialize_b64(self, o, protected):
+        if protected is None:
+            b64n = None
+        else:
+            p = json_decode(protected)
+            b64n = p.get('b64')
+            if b64n is not None:
+                if not isinstance(b64n, bool):
+                    raise InvalidJWSObject('b64 header must be boolean')
+        b64 = o.get('b64')
+        if b64 == b64n:
+            return
+        elif b64 is None:
+            o['b64'] = b64n
+        else:
+            raise InvalidJWSObject('conflicting b64 values')
+
     def deserialize(self, raw_jws, key=None, alg=None):
         """Deserialize a JWS token.
 
@@ -455,25 +376,21 @@ class JWS(object):
         try:
             try:
                 djws = json_decode(raw_jws)
-                o['payload'] = base64url_decode(str(djws['payload']))
                 if 'signatures' in djws:
                     o['signatures'] = list()
                     for s in djws['signatures']:
-                        os = dict()
-                        os['signature'] = base64url_decode(str(s['signature']))
-                        if 'protected' in s:
-                            p = base64url_decode(str(s['protected']))
-                            os['protected'] = p.decode('utf-8')
-                        if 'header' in s:
-                            os['header'] = s['header']
+                        os = self._deserialize_signature(s)
                         o['signatures'].append(os)
+                        self._deserialize_b64(o, os.get('protected'))
                 else:
-                    o['signature'] = base64url_decode(str(djws['signature']))
-                    if 'protected' in djws:
-                        p = base64url_decode(str(djws['protected']))
-                        o['protected'] = p.decode('utf-8')
-                    if 'header' in djws:
-                        o['header'] = djws['header']
+                    o = self._deserialize_signature(djws)
+                    self._deserialize_b64(o, o.get('protected'))
+
+                if 'payload' in djws:
+                    if o.get('b64', True):
+                        o['payload'] = base64url_decode(str(djws['payload']))
+                    else:
+                        o['payload'] = djws['payload']
 
             except ValueError:
                 c = raw_jws.split('.')
@@ -482,6 +399,7 @@ class JWS(object):
                 p = base64url_decode(str(c[0]))
                 if len(p) > 0:
                     o['protected'] = p.decode('utf-8')
+                    self._deserialize_b64(o, o['protected'])
                 o['payload'] = base64url_decode(str(c[1]))
                 o['signature'] = base64url_decode(str(c[2]))
 
@@ -504,7 +422,8 @@ class JWS(object):
         :param potected: The Protected Header (optional)
         :param header: The Unprotected Header (optional)
 
-        :raises InvalidJWSObject: if no payload has been set on the object.
+        :raises InvalidJWSObject: if no payload has been set on the object,
+                                  or invalid headers are provided.
         :raises ValueError: if the key is not a :class:`JWK` object.
         :raises ValueError: if the algorithm is missing or is not provided
          by one of the headers.
@@ -515,16 +434,36 @@ class JWS(object):
         if not self.objects.get('payload', None):
             raise InvalidJWSObject('Missing Payload')
 
+        b64 = True
+
         p = dict()
         if protected:
-            p = json_decode(protected)
-            # TODO: allow caller to specify list of headers it understands
-            if 'crit' in p:
-                self._check_crit(p['crit'])
+            if isinstance(protected, dict):
+                p = protected
+                protected = json_encode(p)
+            else:
+                p = json_decode(protected)
 
+        # If b64 is present we must enforce criticality
+        if 'b64' in list(p.keys()):
+            crit = p.get('crit', [])
+            if 'b64' not in crit:
+                raise InvalidJWSObject('b64 header must always be critical')
+            b64 = p['b64']
+
+        if 'b64' in self.objects:
+            if b64 != self.objects['b64']:
+                raise InvalidJWSObject('Mixed b64 headers on signatures')
+
+        h = None
         if header:
-            h = json_decode(header)
-            p = self._merge_headers(p, h)
+            if isinstance(header, dict):
+                h = header
+                header = json_encode(header)
+            else:
+                h = json_decode(header)
+
+        p = self._merge_check_headers(p, h)
 
         if 'alg' in p:
             if alg is None:
@@ -552,21 +491,18 @@ class JWS(object):
         elif 'signature' in self.objects:
             self.objects['signatures'] = list()
             n = dict()
-            n['signature'] = self.objects['signature']
-            del self.objects['signature']
+            n['signature'] = self.objects.pop('signature')
             if 'protected' in self.objects:
-                n['protected'] = self.objects['protected']
-                del self.objects['protected']
+                n['protected'] = self.objects.pop('protected')
             if 'header' in self.objects:
-                n['header'] = self.objects['header']
-                del self.objects['header']
+                n['header'] = self.objects.pop('header')
             if 'valid' in self.objects:
-                n['valid'] = self.objects['valid']
-                del self.objects['valid']
+                n['valid'] = self.objects.pop('valid')
             self.objects['signatures'].append(n)
             self.objects['signatures'].append(o)
         else:
             self.objects.update(o)
+            self.objects['b64'] = b64
 
     def serialize(self, compact=False):
         """Serializes the object into a JWS token.
@@ -579,7 +515,6 @@ class JWS(object):
         :raises InvalidJWSSignature: if no signature has been added
          to the object, or no valid signature can be found.
         """
-
         if compact:
             if 'signatures' in self.objects:
                 raise InvalidJWSOperation("Can't use compact encoding with "
@@ -592,23 +527,40 @@ class JWS(object):
                 protected = base64url_encode(self.objects['protected'])
             else:
                 protected = ''
-            return '.'.join([protected,
-                             base64url_encode(self.objects['payload']),
+            if self.objects.get('payload', False):
+                if self.objects.get('b64', True):
+                    payload = base64url_encode(self.objects['payload'])
+                else:
+                    if isinstance(self.objects['payload'], bytes):
+                        payload = self.objects['payload'].decode('utf-8')
+                    else:
+                        payload = self.objects['payload']
+                    if '.' in payload:
+                        raise InvalidJWSOperation(
+                            "Can't use compact encoding with unencoded "
+                            "payload that uses the . character")
+            else:
+                payload = ''
+            return '.'.join([protected, payload,
                              base64url_encode(self.objects['signature'])])
         else:
             obj = self.objects
+            sig = dict()
+            if self.objects.get('payload', False):
+                if self.objects.get('b64', True):
+                    sig['payload'] = base64url_encode(self.objects['payload'])
+                else:
+                    sig['payload'] = self.objects['payload']
             if 'signature' in obj:
                 if not obj.get('valid', False):
                     raise InvalidJWSSignature("No valid signature found")
-                sig = {'payload': base64url_encode(obj['payload']),
-                       'signature': base64url_encode(obj['signature'])}
+                sig['signature'] = base64url_encode(obj['signature'])
                 if 'protected' in obj:
                     sig['protected'] = base64url_encode(obj['protected'])
                 if 'header' in obj:
                     sig['header'] = obj['header']
             elif 'signatures' in obj:
-                sig = {'payload': base64url_encode(obj['payload']),
-                       'signatures': list()}
+                sig['signatures'] = list()
                 for o in obj['signatures']:
                     if not o.get('valid', False):
                         continue
@@ -632,24 +584,27 @@ class JWS(object):
             raise InvalidJWSOperation("Payload not verified")
         return self.objects['payload']
 
+    def detach_payload(self):
+        self.objects.pop('payload', None)
+
     @property
     def jose_header(self):
         obj = self.objects
         if 'signature' in obj:
-            jh = dict()
             if 'protected' in obj:
                 p = json_decode(obj['protected'])
-                jh = self._merge_headers(jh, p)
-            jh = self._merge_headers(jh, obj.get('header', dict()))
-            return jh
+            else:
+                p = None
+            return self._merge_check_headers(p, obj.get('header', dict()))
         elif 'signatures' in self.objects:
             jhl = list()
             for o in obj['signatures']:
                 jh = dict()
-                if 'protected' in obj:
+                if 'protected' in o:
                     p = json_decode(o['protected'])
-                    jh = self._merge_headers(jh, p)
-                jh = self._merge_headers(jh, o.get('header', dict()))
+                else:
+                    p = None
+                jh = self._merge_check_headers(p, o.get('header', dict()))
                 jhl.append(jh)
             return jhl
         else:
